////////////////////////////////////
// gllib.h
// OpenGL stuff
// Original coding by Pascal Gautron, July 2006
////////////////////////////////////

#ifndef _GLLIB_H_
#define _GLLIB_H_
// standard headers
#ifdef WIN32
#include <windows.h>
#endif //WIN32
// OpenGL extension header
#include <GL/glew/glew.h>
// WGL OpenGL extension header
#include <GL/glew/wglew.h>
#include <GL/glut.h>
#include <core.h>


inline float InvSqrt (float x)
{
    float xhalf = 0.5f*x;
    int i = *(int*)&x;
    i = 0x5f3759df - (i >> 1);
    x = *(float*)&i;
    x = x*(1.5f - xhalf*x*x);
    return x;
}

// Display text in OpenGl window
/*
Parameters:
  - pos [in] 2D raster position at which text has to be inserted
  - string [in] Text
  - font [in] Optionnal, the font used for the text (default = GLUT_BITMAP_8_BY_13)
*/
inline void BitmapOutput2D(const float pos[2], const char *string) 
{
  int len,i;
  glRasterPos2fv(pos);
  //glutBitmapString (GLUT_BITMAP_8_BY_13, string);
  len = (int) strlen(string);
  for (i = 0; i < len; i++) {
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13,string[i]); }
}


// Load a texture on GPU memory.
/**
\param fileName [in] Image file name
\return OpenGL texture ID if succeeded, 0 otherwise
*/
GLuint LoadTexture(char *fileName);
/*{
 /* if (
    fileName[strlen(fileName)-1]=='/' ||
    fileName[strlen(fileName)-1]=='\\' 
    ) return 0;
  ilInit();
  ilImage img;
  if (!img.Load(fileName))
  {
    WARNING("Texture file " << fileName << " cannot be opened");
    return 0;
  }
  GLuint texID;
  glGenTextures(1, &texID);
  glBindTexture(GL_TEXTURE_2D, texID);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, img.Width(), img.Height(), 0, GL_RGB, GL_UNSIGNED_BYTE, img.GetData());
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  return texID;
}*/

inline bool DisplayGLError(char *glerror)
{
	GLenum err = glGetError();
	switch(err)
	{
	case GL_NONE:
		sprintf(glerror, "GL Error : NONE");
    return false;
		break;
	case GL_INVALID_ENUM:
		sprintf(glerror, "GL Error : GL_INVALID_ENUM");
		break;
	case GL_INVALID_VALUE:
		sprintf(glerror, "GL Error : GL_INVALID_VALUE");
		break;
	case GL_INVALID_OPERATION:
		sprintf(glerror, "GL Error : GL_INVALID_OPERATION");
		break;
	case GL_STACK_OVERFLOW:
		sprintf(glerror, "GL Error : GL_STACK_OVERFLOW");
		break;
	case GL_STACK_UNDERFLOW:
		sprintf(glerror, "GL Error : GL_STACK_UNDERFLOW");
		break;
	case GL_OUT_OF_MEMORY:
		sprintf(glerror, "GL Error : GL_OUT_OF_MEMORY");
		break;
	default:
		sprintf(glerror, "GL Error : Unknown (0x%x)", err);
	}
  return true;
}


inline bool DisplayGLError()
{
  char buffer[64];
  bool res = DisplayGLError(buffer);
  if (res)
    INFO(buffer);
  return res;
}


  // 3D TEXTURES
  // -----------
  ///\name 3D Textures management and filling
  ///\{
  ///\brief 3D Texture representation
  template <class T> class C3DTexture
  {
  public:
    /// Constructor
    C3DTexture()
      :_width(0), _height(0), _depth(0), _cPP(0), _texData(NULL) {};
    /// Initializer
    /**
    \param width [in] Texture width
    \param height [in] Texture height
    \param depth [in] Texture depth
    \param componentsPerPixel [in] Components per pixel (3 for RGB)
    */
    void Init(
      unsigned int width, 
      unsigned int height, 
      unsigned int depth,
      unsigned int componentsPerPixel)
    {
      if (_texData) 
      {
        WARNING("Initializing already initialized 3D texture");
        delete[] _texData;
      }
      _width   = width;
       _height  = height;
       _depth   = depth;
       _cPP     = componentsPerPixel;
       _texData = new T[width*height*depth*componentsPerPixel];
       if (_texData == NULL)
       {
         FATAL("Cannot allocate 3D Texture memory");
       }
    }

    /// Destroy texture data
    void Erase() 
    {
      if (_texData) delete[] _texData;
      _texData = NULL;
      _width = _height = _depth = _cPP = 0;
    }

    /// Destructor
    virtual ~C3DTexture()
    {
      if (_texData)
        delete[] _texData;
    }

    /// Pixel assignment
    /**
    \param x [in] X position in texture
    \param y [in] Y position in texture
    \param z [in] Z position in texture
    \param data [in] Pixel data
    */
    void SetPixel(
      unsigned int x, 
      unsigned int y, 
      unsigned int z, 
      T *data)
    {
      /*if (x>=_width || y>=_height || z>=_depth)
      {
        WARNING("SetPixel: trying to assign invalid position in 3D texture");
        return;
      }*/
      memcpy(_texData + _cPP*(_width*(y+_height*z)+x), data, _cPP*sizeof(T));
    }

    T* GetPixel(
      unsigned int x, 
      unsigned int y, 
      unsigned int z)
    {
      /*if (x>=_width || y>=_height || z>=_depth)
      {
        WARNING("GetPixel: trying to get invalid position in 3D texture");
        return NULL;
      }*/
      return (_texData + _cPP*(_width*(y+_height*z)+x));
    }

    /// Texture Z slice assignment
    /**
    \param z [in] Z position of the slice
    \param data [in] Pixel data for the slice
    */
    void SetZSlice(
      unsigned int z,
      T *data
      )
    {
      if (z>=_depth)
      {
        WARNING("SetZSlice: trying to assign invalid position in 3D texture");
        return;
      }
      memcpy(_texData + _cPP*(_width*_height*z), data, _cPP*sizeof(T)*(_width*_height));
    }

    /// Texture Z slice accessor
    /**
    \param z [in] Z position of the slice
    */
    T *GetZSlice(
      unsigned int z
      )
    {
      if (z>=_depth)
      {
        WARNING("SetZSlice: trying to access invalid position in 3D texture");
        return NULL;
      }
      return _texData + _cPP*(_width*_height*z);
    }

    /// Indicates whether the texture has allocated memory
    /**
    \return True if the data pointer is not null
    */
    bool IsAllocated() const {return (_texData!=NULL);}
    /// Texture data accessor
    /**
    \return Pointer on texture data, ready for glTexImage3D
    */
    T *GetData() {return _texData;}
  private:
    /// Texture width
    unsigned int _width;
    /// Texture height
    unsigned int _height;
    /// Texture depth
    unsigned int _depth;
    /// Number of components per pixel
    unsigned int _cPP;

    /// Texture data
    T *_texData;
  };
  ///\}

	// PROGRAMMABLE SHADERS
	// --------------------
	///\name Programmable shaders
	///\{
	/// Test for programmable shaders capability - display an error message if necessary
	/**
	Currently just tests for hardware capabilities
	\return True if hardware is capable of programmable shading
	*/
	inline bool InitProgShaders()
  {
    if (!GLEW_ARB_shader_objects)
		  WARNING("GLEW_ARB_shader_objects extension not available");
	  if (!GLEW_ARB_shading_language_100)
		  WARNING("GLEW_ARB_shading_language_100 extension not available");
	  if (!GLEW_ARB_vertex_shader)
		  WARNING("GLEW_ARB_vertex_shader extension not available");
	  if (!GLEW_ARB_fragment_shader)
		  WARNING("GLEW_ARB_fragment_shader extension not available");
	  return GLEW_ARB_shader_objects&&GLEW_ARB_shading_language_100&&GLEW_ARB_vertex_shader&&GLEW_ARB_fragment_shader;
  }

  /// Shader programs wrapper
  class CShader
  {
  public:
    void GetCompilationInfo(GLenum shader, char *info, int maxLength, int *infoLength) const
    {
      static char *buffer = new char[1024*1024];
      int len;
      int counter=0;
      bool finished=false;
	    glGetInfoLogARB(shader, 1024*1024, &len, buffer);
      char *line = strtok(buffer, "\n");
      while (line!=NULL && !finished)
      {
        int linelength = (int)strlen(line);
        // The error line is not a 'warning'
        if (
          strstr(line, "warning C")==NULL && 
          strstr(line, "-----")==NULL &&
          strstr(line, "Vertex info")==NULL &&
          strstr(line, "Fragment info")==NULL)
        {
          if (counter+linelength < maxLength-1)
          {
            sprintf(info+counter, "%s\n", line);
            //memcpy(info+counter, line, linelength);
            counter+=linelength;
          }
          else
          {
            finished = true;
          }
        }
        line = strtok(NULL, "\n");
      }
      if (counter)
      {
        info[counter+1] = '\0';
      }
      else
      { 
        info[0] = '\0';
      }
      *infoLength = counter;
    }

    // Compile a shader source from a file
    /*
    Parameters:
	    - filename [in] Name of the file containing the shader source
	    - shader [in, out] Shader identifier, obtained using Create{Vertex, Fragment}Shader
    */
    void CompileShaderFromFile(const char *filename, GLenum shader)
    {
	    FILE *file = NULL;
      file = fopen(filename, "rb");
	    if (file==NULL)
	    {
		    WARNING("Cannot open shader file: " << filename);
		    return;
	    }
	    fseek(file, 0, SEEK_END);
	    long filesize = ftell(file);
	    fseek(file, 0, SEEK_SET);
	    char *source = new char[filesize+1];
	    fread(source, 1, filesize*sizeof(char), file);
	    fclose(file);
	    source[filesize] = '\0';
      const char *csrc = source;
	    glShaderSourceARB(shader, 1, &csrc, NULL);
	    glCompileShaderARB(shader);

	    //CompileShaderFromString(source, shader);
	    delete[] source;

      char errmsg[8192];
      int l=0;
      GetCompilationInfo(shader, errmsg, 8192, &l);
      if (l>0)
      {
        WARNING("Shader " << filename << " compilation error:");
        WARNING(errmsg);
      }

    }

    // Link vertex and fragment shader together in a program
    /*
    Parameters:
	    - vshader [in] Vertex shader identifier
	    - fshader [in] Fragment shader identifier
	    - program [in, out] Program object
    */
    void LinkShaders(GLenum vshader, GLenum fshader, GLenum program)
    {
      char glerror[64];
      if (DisplayGLError(glerror))
      {
        INFO("Error before attach vertex shader " << glerror);
      }
      glAttachObjectARB(program, vshader);
      if (DisplayGLError(glerror))
      {
        INFO("Failed to attach vertex shader " << glerror);
      }
	    glAttachObjectARB(program, fshader);
      if (DisplayGLError(glerror))
      {
        INFO("Failed to attach fragment shader " << glerror);
      }

      glLinkProgramARB(program);
        if (DisplayGLError(glerror))
      {
        INFO("Failed to link shaders " << glerror);
      }

    }

    /// Constructor
    CShader()
    :_hasVS(false), _hasFS(false), _hasPgm(false)
    {
      _vs = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
      _fs = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
      _pgm = glCreateProgramObjectARB();
    }
    /// Copy constructor
    /**
    WARNING : only copies the shaders IDs. Shader is not duplicated in GPU memory
    */
    CShader(const CShader &source)
    :
      _vs(source._vs),
      _fs(source._fs),
      _pgm(source._pgm),
      _hasVS(source._hasVS),
      _hasFS(source._hasFS),
      _hasPgm(source._hasPgm)

    {
      strcpy(_name, source._name);
    }
    /// Constructor
    /**
    \param vs [in] Vertex shader source file
    \param fs [in] Fragment shader source file
    */
    CShader(const char *vs, const char *fs)
    {
      _vs = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
      _fs = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
      _pgm = glCreateProgramObjectARB();
      CompileShaderFromFile(vs, _vs);
      _hasVS = true;
      CompileShaderFromFile(fs, _fs);
      _hasFS = true;
      LinkShaders(_vs, _fs, _pgm);
      _hasPgm = true;

      int i=0, j=0;
      for(i=(int)strlen(vs)-1; i > 0 && vs[i] != '/'; i--);
      i++;
      for(; vs[i] != '.' && i < 511; i++, j++)
      {
        _name[j] = vs[i];
      }
      _name[j] = '\0';

      char tmp[10000];
      int l;
      
      GetCompilationInfo(_pgm, tmp, 10000, &l);
    /*  glGetInfoLogARB(
                  _pgm,
                  10000,
                  &l,
                  tmp);*/
      if (strlen(tmp)>0)
      {
        INFO("\"" << _name << "\" shader link log: ");
        INFO(tmp);
      }

    }
    /// Destructor
    virtual ~CShader()
    {
      glDeleteObjectARB(_vs);
      glDeleteObjectARB(_fs);
      glDeleteObjectARB(_pgm);
    }
    /// Load vertex shader source code
    void LoadVS(const char *vs)
    {
       CompileShaderFromFile(vs, _vs);
      _hasVS = true;
      _hasPgm = false;
      int i=0, j=0;
      for(i=(int)strlen(vs)-1; i > 0 && vs[i] != '/'; i--);
      i++;
      for(; vs[i] != '.' && i < 511; i++, j++)
      {
        _name[j] = vs[i];
      }
      _name[j] = '\0';
    }
    /// Load fragment shader source code
    void LoadFS(const char *fs)
    {
      CompileShaderFromFile(fs, _fs);
      _hasFS = true;
      _hasPgm = false;
    }
	/// Display compilation info log
	void DisplayInfoLog() const
  {
    char tmp[10000];
    int l;
    
    GetCompilationInfo(_pgm, tmp, 10000, &l);
    /*glGetInfoLogARB(
                _pgm,
                10000,
                &l,
                tmp);*/
    if (l>0)
    {
      INFO("\"" << _name << "\" shader link log: ");
      INFO(tmp);
    }
  }

	/// Get compilation info log
	/**
	\param log [out] Info log
	*/
	void DisplayInfoLog(char *log) const
  {
    int l;
    GetCompilationInfo(_pgm, log, 10000, &l);
  /*  glGetInfoLogARB(
                _pgm,
                10000,
                &l,
                log);*/
  }

    /// Link shaders
    void Link()
    {
      if (!_hasVS || !_hasFS)
      {
        WARNING("Cannot link shader program: vertex or fragment shader missing");
	    return;
      }

      if (!_hasPgm)
      {
        LinkShaders(_vs, _fs, _pgm);
        _hasPgm = true;
      }
    }

    // Use program for vertex and fragment processing
    /*
    Put program=0 to use classical 'fixed-function' processing
    Parameters:
	    - program [in] Program to be used in vertex and fragment processing
    */
    void UseProgram(GLenum program)
    {
      if (DisplayGLError())
      {
        WARNING("UseProgram: a previous GL Error detected");
      }
	    glUseProgramObjectARB(program);
      char glerror[64];
      if (DisplayGLError(glerror))
      {
        WARNING("Cannot use program " << program << "(" << glerror << ")");
        glUseProgramObjectARB(0);
      }
    }

    /// Activate shader program
    void Activate()
    {
      if (_hasPgm)
      {
    //    GLFUNCLIB_INFO("Activating shader " << _name);
        UseProgram(_pgm);
      }
      else
        WARNING("Cannot use shader program: program not linked");
    }

    /// Try to validate the shader and its uniform values
    void Validate() {glValidateProgramARB(_pgm);}

    /// Set float uniform
    /**
    \param name [in] Variable name
    \param n [in] Number of values in one variable
    \param count [in] Number of variables
    \param v [in] Values
    */
    void SetFloatUniform(
      const char *name,
      GLsizei n,
      GLsizei count,
      GLfloat *v)
    {
      if (!_hasPgm)
      {
        WARNING("Trying to set uniform float to invalid shader " << _name);
        DisplayGLError();
        return;
      }
      GLint id;
      id = glGetUniformLocationARB(_pgm, name);
      if (id==-1) 
      {
        WARNING("Uniform float variable " << name << " not found in " << _name);
        DisplayGLError();
        return;
      }
      
      switch(n)
      {
      case 1:
        glUniform1fvARB(id, count, v);
        break;
      case 2:
        glUniform2fvARB(id, count, v);
        break;
      case 3:
        glUniform3fvARB(id, count, v);
        break;
      case 4:
        glUniform4fvARB(id, count, v);
        break;
      default:
        WARNING("Trying to set an invalid float array in "  << _name << " (length=" << n << ")");
      }
    }

    /// Set float single-valued uniform
    /**
    \param name [in] Variable name
    \param v [in] Value
    */
    void SetFloatUniform(
      const char *name,
      GLfloat v)
      {
        if (!_hasPgm)
        {
          WARNING("Trying to set uniform float to invalid shader " << _name);
          DisplayGLError();
          return;
        }
        GLint id;
        id = glGetUniformLocationARB(_pgm, name);
        if (id==-1) 
        {
          //GLFUNCLIB_WARNING("Uniform float variable " << name << " not found in " << _name);
          DisplayGLError();
          return;
        }
        glUniform1fARB(id, v);
      }

    /// Set int single-valued uniform
    /**
    \param name [in] Variable name
    \param v [in] Value
    */
    void SetIntUniform(
      const char *name,
      GLint v)
    {
      if (!_hasPgm)
      {
        WARNING("Trying to set uniform float to invalid shader " << _name);
        DisplayGLError();
        return;
      }
      GLint id;
      id = glGetUniformLocationARB(_pgm, name);
      if (id==-1) 
      {
        //GLFUNCLIB_WARNING("Uniform int variable " << name << " not found in " << _name);
        DisplayGLError();
        return;
      }
      glUniform1iARB(id, v);
    }


    /// Set integer uniform
    /**
    \param name [in] Variable name
    \param n [in] Number of values in one variable
    \param count [in] Number of variables
    \param v [in] Values
    */
    void SetIntUniform(
      const char *name,
      GLsizei n,
      GLsizei count,
      GLint *v)
    {

      if (!_hasPgm)
      {
        WARNING("Trying to set uniform int to invalid shader " << _name);
        DisplayGLError();
        return;
      }

      GLint id;
      id = glGetUniformLocationARB(_pgm, name);
      if (id==-1) 
      {
        //GLFUNCLIB_WARNING("Uniform int variable " << name << " not found in " << _name);
        DisplayGLError();
        return;
      }
      
      switch(n)
      {
      case 1:
        glUniform1ivARB(id, count, v);
        break;
      case 2:
        glUniform2ivARB(id, count, v);
        break;
      case 3:
        glUniform3ivARB(id, count, v);
        break;
      case 4:
        glUniform4ivARB(id, count, v);
        break;
      default:
        WARNING("Trying to set an invalid int array in "  << _name << " (length=" << n << ")");
      }
    }

    /// Set matrix uniform
    /**
    \param name [in] Variable name
    \param n [in] Matrix size (2, 3 or 4)
    \param count [in] Number of matrices
    \param v [in] Values
    \param transpose [in] If true, matrices will be transposed before storage
    */
    void SetMatrixUniform(
      const char *name,
      GLsizei n,
      GLsizei count,
      GLfloat *v,
      GLboolean transpose
      )
    {
      if (!_hasPgm)
      {
        WARNING("Trying to set uniform matrix to invalid shader " << _name);
        DisplayGLError();
        return;
      }
      GLint id;
      id = glGetUniformLocationARB(_pgm, name);
      if (id==-1) 
      {
        WARNING("Uniform matrix variable " << name << " not found in " << _name);
        DisplayGLError();
        return;
      }
      
      switch(n)
      {
      case 2:
        glUniformMatrix2fvARB(id, count, transpose, v);
        break;
      case 3:
        glUniformMatrix3fvARB(id, count, transpose, v);
        break;
      case 4:
        glUniformMatrix4fvARB(id, count, transpose, v);
        break;
      default:
        WARNING("Trying to set an invalid matrix in "  << _name << " (length=" << n << ")");
      }
    }

    /// Bind attribute location
    /**
      Attribute location binding MUST BE DONE BEFORE SHADER LINKING
      \param index [in] Location index to bind
      \param name [in] Identifier in vertex shader for this location
    */
    void BindAttributeLocation(GLuint index, const char *name) {if (_hasVS) glBindAttribLocationARB(_pgm, index, name);}

  private:
    /// Vertex shader
    GLenum _vs;
    /// Fragment shader
    GLenum _fs;
    /// Shader program
    GLenum _pgm;
    /// True if a valid vertex shader is loaded
    bool _hasVS;
    /// True if a valid fragment shader is loaded
    bool _hasFS;
    /// True if the shader program is linked and usable
    bool _hasPgm;
    /// Shader name
    char _name[512];
  };

  // FRAME BUFFER OBJECT
  // -------------------
  ///\name Frame Buffer Objects: a nice replacement for PBuffers

class CFBO
{
public:
  CFBO(unsigned int resX, unsigned int resY, unsigned int fboCount, bool depthBuffer)
    :_resX(resX), _resY(resY), _fboCount(fboCount), _texID(new GLuint[fboCount])
  {
	  glGenFramebuffersEXT(1, &_fb);
    glGenTextures(fboCount, _texID);
    if (depthBuffer)
    {
      Bind();
      //glGenTextures(1, &_depth);
      // Setup depth_stencil texture (not mipmap)
      /*glBindTexture(GL_TEXTURE_2D, _depth);
      GLFuncLib::DisplayGLError();
      glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT24,
                    resX, resY, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
      GLFuncLib::DisplayGLError();
      glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
                                GL_DEPTH_ATTACHMENT_EXT,
                                GL_TEXTURE_2D, _depth, 0);
      GLFuncLib::DisplayGLError();*/
/*      glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
                                GL_STENCIL_ATTACHMENT_EXT,
                                GL_TEXTURE_2D, _depth, 0);
      GLFuncLib::DisplayGLError();*/

	    glGenRenderbuffersEXT(1, &_depth);    
      glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, _depth);
      glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_STENCIL_EXT, resX, resY);
      glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
                                GL_RENDERBUFFER_EXT, _depth);
      glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_STENCIL_ATTACHMENT_EXT,
                                GL_RENDERBUFFER_EXT, _depth);
    }
    else
      _depth = 0;

    
  }
  virtual ~CFBO()
  {
	  glDeleteTextures(4, _texID);
	  glDeleteFramebuffersEXT(1, &_fb);
	  glDeleteRenderbuffersEXT(1, &_depth);
  }
  void Bind() {glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _fb);}
  void UnBind() {glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);}
  GLuint GetTexture(unsigned int index)
  {
    if (index < _fboCount)
    {
      return _texID[index];
    }
    else
    {
      std::cerr << "Bad texture index for FBO (index=" << index << ")" << std::endl;
      return 0;
    }    
  }
  void AttachTexture2D(unsigned int index)
  {
    if (index < _fboCount)
    {
      glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT+index, GL_TEXTURE_2D, _texID[index], 0);
    }
    else
    {
      std::cerr << "Bad texture index for FBO (index=" << index << ")"<< std::endl;
    }
  }

  void AttachTexture3D(unsigned int index, GLint texID, unsigned int zOffset)
  {
    if (index < _fboCount)
    {
      glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT+index, GL_TEXTURE_3D, texID, 0, zOffset);
    }
    else
    {
      std::cerr << "Bad texture index for FBO (index=" << index << ")"<< std::endl;
    }
  }

  void SetDrawBuffer(unsigned int index)
  {
    if (index < _fboCount)
    {
      glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT+index);
    }
    else
    {
      std::cerr << "Bad draw buffer index for FBO (index=" << index << ")"<< std::endl;
    }
  }
  void SetDrawBuffers(unsigned int count, unsigned int *index)
  {
    GLuint bufs[4];
    for (unsigned int i=0; i < count; i++)
    {
      if (index[i] >= _fboCount)
      {
        std::cerr << "Bad draw buffer index for FBO (index=" << index[i] << ")"<< std::endl;
        return;
      }
      bufs[i] = GL_COLOR_ATTACHMENT0_EXT+index[i];
    }
    glDrawBuffersARB(count, bufs);
  }


  void CheckStatus()
  {
    GLenum status;
    status = (GLenum) glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    switch(status) {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
            break;
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
          std::cerr << "Unsupported framebuffer format" << std::endl;
            break;
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
            std::cerr << "Framebuffer incomplete, missing attachment" << std::endl;
            break;
//        case GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT:
//            WARNING("Framebuffer incomplete, duplicate attachment");
//            break;
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
            std::cerr << "Framebuffer incomplete, attached images must have same dimensions" << std::endl;
            break;
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
            std::cerr << "Framebuffer incomplete, attached images must have same format" << std::endl;
            break;
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
            std::cerr << "Framebuffer incomplete, missing draw buffer" << std::endl;
            break;
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
            std::cerr << "Framebuffer incomplete, missing read buffer" << std::endl;
            break;
        default:
            std::cerr << "Unknown framebuffer error" << std::endl;
    }
  }
private:
  GLuint _fb;
  GLuint _depth;
  unsigned int _resX, _resY;
  unsigned int _fboCount;
  GLuint *_texID;
};


#endif //_GLLIB_H_