////////////////////////////////////
// matrix.h
// Matrix operations
// Original coding by Pascal Gautron, July 2006
////////////////////////////////////


// Standard headers
#include <cmath>
#include <float.h>
#include <iostream>
#include <cassert>

#ifndef __MATRIX_H__
#define __MATRIX_H__

// 4x4 MATRICES OPERATIONS
	// -----------------------

	/// \name 4x4 Matrices operations
	/// \{

	/** \brief Operations on 4x4 matrices*/
	namespace Matrix4x4
	{
  inline float &Elem0(float m[16], int row, int col)
  {
    return m[col*4+row];
  }
  inline float &Elem1(float m[16], int row, int col)
  {
    return m[(col-1)*4+row-1];
  }
  /// Matrix assignment
		/**
		\param src [in] Source matrix
		\param dst [out] Destination matrix
		\return Pointer to 'dst'
		*/
		inline float *Assign(const float src[16], float dst[16]) {memcpy(dst, src, 16*sizeof(float)); return dst;}

		/// Matrix component-wise add, in place
		/** Equivalent to m2 += k*m1
		*/
		inline float *AddIP(float k, const float m1[16], float m2[16]) {unsigned int i; for(i=0; i < 16; i++) {m2[i] += k*m1[i];} return m2;}


		/// Vector-Matrix multiplication
		/** Multiply the LINE vector 'v' by matrix 'm' 
		\param v [in] Vector
		\param m [in] Matrix
		\param dest [out] v*m
		*/
		inline void Mult(const float v[3], const float m[16], float dest[3]) {dest[0] = (v[0]*m[0] + v[1]*m[1] + v[2]*m[2]);  dest[1] = (v[0]*m[4] + v[1]*m[5] + v[2]*m[6]);  dest[2] = (v[0]*m[8] + v[1]*m[9] + v[2]*m[10]); }


		/// Multiplication by a scalar, in place
		/**
		Equivalent to m *= k;
		\param k [in] Factor
		\param m [in, out] Multiplied matrix
		*/
		inline void MultIP(float k, float m[16]) {for(unsigned int i=0; i < 16; i++) m[i] *= k;}

		/// Vector-Matrix multiplication
		/** Multiply the LINE vector 'v' by matrix 'm' 
		\param v [in] Vector
		\param m [in] Matrix
		\param dest [out] v*m
		*/
		inline void Mult(const float v[3], const double m[16], float dest[3]) {dest[0] = (v[0]*(float)m[0] + v[1]*(float)m[1] + v[2]*(float)m[2]);  dest[1] = (v[0]*(float)m[4] + v[1]*(float)m[5] + v[2]*(float)m[6]);  dest[2] = (v[0]*(float)m[8] + v[1]*(float)m[9] + v[2]*(float)m[10]); }

		/// Vector-Matrix multiplication
		/** Multiply the LINE vector 'v' by matrix 'm' 
		\param v [in] Vector
		\param m [in] Matrix
		\param dest [out] v*m
		*/
		inline void Mult(const double v[3], const double m[16], float dest[3]) {double w = 1.0/(v[0]*m[12] + v[1]*m[13] + v[2]*m[14] + m[15]); dest[0] = (float)((v[0]*m[0] + v[1]*m[1] + v[2]*m[2]+m[3])*w);  dest[1] = (float)((v[0]*m[4] + v[1]*m[5] + v[2]*m[6]+m[7])*w);  dest[2] = (float)((v[0]*m[8] + v[1]*m[9] + v[2]*m[10]+m[11])*w); }

		/// Matrix-vector multiplication
		/** Multiply the matrix 'm' by the COLUMN vector 'v'
		\param m [in] Matrix
		\param v [in] Vector
		\param dest [out] m*v
		*/
	//	inline void MultCol(const float m[16], const float v[3], float dest[3]) {dest[0] = (v[0]*m[0] + v[1]*m[4] + v[2]*m[8]);  dest[1] = (v[0]*m[1] + v[1]*m[5] + v[2]*m[9]);  dest[2] = (v[0]*m[2] + v[1]*m[6] + v[2]*m[10]); }
		inline void MultCol(const float m[16], const float v[3], float dest[3]) {
			int i;
			float in[4];
			//memcpy(in, v, 3*sizeof(float));
			in[3] = 1.0f;
			for (i=0; i<4; i++) {
			in[i] = 
				v[0] * m[0*4+i] +
				v[1] * m[1*4+i] +
				v[2] * m[2*4+i] +
				1.0f * m[3*4+i];
			}
			dest[0] = in[0]/in[3];
			dest[1] = in[1]/in[3];
			dest[2] = in[2]/in[3];
		}

    inline void ModelViewProd(const float m[16], const float v[3], float dest[3])
    {
      dest[0] = v[0]*m[0]+v[1]*m[4]+v[2]*m[8] + m[12];
      dest[1] = v[0]*m[1]+v[1]*m[5]+v[2]*m[9] + m[13];
      dest[2] = v[0]*m[2]+v[1]*m[6]+v[2]*m[10]+ m[14];
    }

		inline void MultCol3x3(const float m[16], const float v[3], float dest[3]) {
  

      int i;
			float in[3];
			//memcpy(in, v, 3*sizeof(float));
			//in[3] = 1.0f;
			for (i=0; i<3; i++) {
			in[i] = 
				v[0] * m[0*4+i] +
				v[1] * m[1*4+i] +
				v[2] * m[2*4+i];
			}
			dest[0] = in[0]; ///in[3];
			dest[1] = in[1]; ///in[3];
			dest[2] = in[2]; ///in[3];
		}


		/// Matrix-vector multiplication
		/** Multiply the matrix 'm' by the COLUMN vector 'v'
		\param m [in] Matrix
		\param v [in] Vector
		\param dest [out] m*v
		*/
		inline void MultCol(const double m[16], const float v[3], float dest[3]) {float w = (v[0]*(float)m[3] + v[1]*(float)m[7] + v[2]*(float)m[11]+(float)m[15]); dest[0] = (v[0]*(float)m[0] + v[1]*(float)m[4] + v[2]*(float)m[8]+(float)m[12])/w;  dest[1] = (v[0]*(float)m[1] + v[1]*(float)m[5] + v[2]*(float)m[9]+(float)m[13])/w;  dest[2] = (v[0]*(float)m[2] + v[1]*(float)m[6] + v[2]*(float)m[10]+(float)m[14])/w; }

		/// Rotation matrix generation
		/** Create a rotation matrix: rotation around 'axis' of angle 'angle'
		\param axis [in] Rotation axis
		\param angle [in] Rotation angle
		\param dest [out] Corresponding rotation matrix
		*/
		inline void Rotation(const float axis[3], float angle, float dest[16]) 
    {	
      float c = cosf(angle), s=sinf(angle), x = axis[0], y=axis[1], z=axis[2];	
/*      dest[0] = x*x+c*(1-x*x);	
      dest[1] = x*y*(1-c)-z*s;	
      dest[2] = x*z*(1-c)+y*s;	
      dest[3] =  0.0f;	
      
      dest[4] = x*y*(1-c)+z*s;	
      dest[5] = y*y+c*(1-y*y);	
      dest[6] = y*z*(1-c)+x*s;	
      dest[7] = 0.0f;	
      
      dest[8] = x*z*(1-c)-y*s;	
      dest[9] = y*z*(1-c)+y*s;	
      dest[10] = z*z+c*(1-z*z);	
      dest[11] = 0.0f;	

      dest[12] = 0.0f;	
      dest[13] = 0.0f;	
      dest[14] = 0.0f;	
      dest[15] = 1.0f;	

*/
      dest[0] = x*x*(1-c)+c;	
      dest[1] = x*y*(1-c)+z*s;	
      dest[2] = x*z*(1-c)-y*s;	
      dest[3] =  0.0f;	
      
      dest[4] = x*y*(1-c)-z*s;	
      dest[5] = y*y*(1-c)+c;	
      dest[6] = y*z*(1-c)+x*s;	
      dest[7] = 0.0f;	
      
      dest[8] = x*z*(1-c)+y*s;	
      dest[9] = y*z*(1-c)-x*s;	
      dest[10] = z*z*(1-c)+c;	
      dest[11] = 0.0f;	
      
      dest[12] = 0.0f;	
      dest[13] = 0.0f;	
      dest[14] = 0.0f;	
      dest[15] = 1.0f;	

    }
		/// Identity matrix generation
		/** 
		\param dest [out] Identity matrix
		*/
		inline void Identity(float dest[16]) {	dest[1] = dest[2] = dest[3] = dest[4] = dest[6] = dest[7] = dest[8] = dest[9] = dest[11] = dest[12] = dest[13] = dest[14] = 0.0f; dest[0] = dest[5] = dest[10] = dest[15] = 1.0f;}

		/// Zero matrix
		/**
		\param dest [out] Zero matrix
		*/
		inline void Zero(float dest[16]) {memset(dest, 0, 16*sizeof(float));}


    /// Right-handed perspective transform matrix generation
 /*   inline void PerspectiveRH(float fovy, float aspect, float znear, float zfar, float dest[16])
    {
      Zero(dest);
      float h = 1.0f/tanf(fovy);
      float w = h/aspect;
      dest[0] = w;
      dest[5] = h;
      dest[10] = zfar/(znear-zfar);
      dest[11] = -1.0f;
      dest[14] = znear*zfar/(znear-zfar);
    }*/
    inline void Perspective(float fovy, float aspect, float zNear, float zFar, float dest[16])
    {
      float sine, cotangent, deltaZ;
      float radians = fovy / 2.0f * M_PI / 180.0f;

      deltaZ = zFar - zNear;
      sine = sin(radians);
      if ((deltaZ == 0) || (sine == 0) || (aspect == 0)) {
        std::cerr << ("Bad parameters to PerspectiveLH") << std::endl;
	      return;
      }
      cotangent = cos(radians) / sine;

      Identity(dest);

      dest[0] = cotangent / aspect;
      dest[5] = cotangent;
      dest[10] = -(zFar + zNear) / deltaZ;
      dest[11] = -1;
      dest[14] = -2 * zNear * zFar / deltaZ;
      dest[15] = 0;

/*
      m[0][0] = cotangent / aspect;
      m[1][1] = cotangent;
      m[2][2] = -(zFar + zNear) / deltaZ;
      m[2][3] = -1;
      m[3][2] = -2 * zNear * zFar / deltaZ;
      m[3][3] = 0;*/
    }

		/// Translation matrix generation
		/** Create a translation matrix of vector v
		\param v [in] Translation vector
		\param dest [out] Generated matrix
		*/
		inline void Translation(const float v[3], float dest[16]) {Identity(dest); dest[12] = v[0]; dest[13] = v[1]; dest[14] = v[2];}




		/// Scaling matrix generation
		/** Create a scaling matrix of ratios v
		\param v [in] Scaling factors
		\param dest [out] Generated matrix
		*/
		inline void Scale(const float sf[3], float dest[16]) {Identity(dest); dest[0] = sf[0]; dest[5] = sf[1]; dest[10] = sf[2];}

    
    /// Matrices line column dot product
		/** Dot product between 'm1' line 'l' and 'm2' column 'c'
		\param m1 [in] First matrix
		\param l [in] Line in 'm1'
		\param m2 [in] Second matrix
		\param c [in] column in 'm2'
		\return Dot product between 'm1' line 'l' and 'm2' column 'c'
		*/
		inline float LineColProd(const float m1[16], unsigned int l, const float m2[16], unsigned int c) { unsigned int tmp = 4*c; return m1[l]*m2[tmp] + m1[l+4]*m2[tmp+1] + m1[l+8]*m2[tmp+2] + m1[l+12]*m2[tmp+3]; }
		/// Matrix multiplication
		/**
		\param m1 [in] First matrix
		\param m2 [in] Second matrix
		\param dest [out] m1*m2
		*/
	//	inline void Product(const float m1[16], const float m2[16], float dest[16]) {dest[0] = LineColProd(m1, 0, m2, 0); dest[1] = LineColProd(m1, 1, m2, 0);dest[2] = LineColProd(m1, 2, m2, 0);dest[3] = LineColProd(m1, 3, m2, 0);dest[4] = LineColProd(m1, 0, m2, 1);dest[5] = LineColProd(m1, 1, m2, 1);dest[6] = LineColProd(m1, 2, m2, 1);dest[7] = LineColProd(m1, 3, m2, 1);dest[8] = LineColProd(m1, 0, m2, 2);dest[9] = LineColProd(m1, 1, m2, 2);dest[10] = LineColProd(m1, 2, m2, 2);dest[11] = LineColProd(m1, 3, m2, 2);dest[12] = LineColProd(m1, 0, m2, 3);dest[13] = LineColProd(m1, 1, m2, 3); dest[14] = LineColProd(m1, 2, m2, 3); dest[15] = LineColProd(m1, 3, m2, 3); }
inline void Product(const float m1[16], const float m2[16], float dest[16]) {
		int i, j;

    for (i = 0; i < 4; i++) {
	for (j = 0; j < 4; j++) {
	    dest[i*4+j] = 
		m1[i*4+0]*m2[0*4+j] +
		m1[i*4+1]*m2[1*4+j] +
		m1[i*4+2]*m2[2*4+j] +
		m1[i*4+3]*m2[3*4+j];
	}
    }
}


    inline void 
    MultCol32(const float m[16], const float v[3], float dest[2]) 
    {
      float t0 = v[0] * m[0*4+0] + v[1] * m[1*4+0] + v[2] * m[2*4+0] + m[3*4+0];
      float t1 = v[0] * m[0*4+1] + v[1] * m[1*4+1] + v[2] * m[2*4+1] + m[3*4+1];
      float id = 1.0f/(v[0] * m[0*4+3] + v[1] * m[1*4+3] + v[2] * m[2*4+3] + m[3*4+3]);
      dest[0] = t0*id;
      dest[1] = t1*id;
    }
		/// Matrix transposition
		/**
		\param m [in] Original matrix
		\param dest [out] Transposed matrix
		*/
		inline void Transpose(const float m[16], float dest[16]) {dest[0] = m[0]; dest[1] = m[4]; dest[2] = m[8];dest[3] = m[12];dest[4] = m[1];dest[5] = m[5];dest[6] = m[9];dest[10] = m[10];dest[11] = m[14];dest[12] = m[3];dest[13] = m[7];dest[14] = m[11];dest[15] = m[15];}

    inline void Swap(float &a, float &b)
    {
      float c = a;
      a = b;
      b = c;
    }
		/// Matrix transposition in place
		/**
		\param m [in, out] Matrix to transpose
		*/
		inline void TransposeIP(float m[16]) {Swap(m[1], m[4]); Swap(m[2], m[8]); Swap(m[3], m[12]);Swap(m[6], m[9]); Swap(m[11], m[14]); Swap(m[7], m[13]);}

		/// Identity testing
		/** Indicates whether the matrix is the identity matrix
		\param m [in] Matrix to test for identity
		\return true is m is the identity matrix, false otherwise
		*/
		inline bool IsIdentity(const float m[16]) {	bool tmp=true; tmp = (m[1] == 0.0f && m[2] == 0.0f && m[3] == 0.0f && m[4] ==0.0f && m[6] == 0.0f && m[7] == 0.0f);  tmp = tmp && m[8] == 0.0f && m[9] == 0.0f && m[11] == 0.0f && m[12] == 0.0f && m[13] == 0.0f && m[14] == 0.0f && m[0] == 1.0f && m[5] == 1.0f && m[10] == 1.0f && m[15] == 1.0f; return tmp;}


		/// Matrix inversion
		/**
		\param m [in] Matrix to invert
		\param dest [out] Inverted matrix
		\return True if inversion has succeeded
		*/
		inline bool Invert(const float m[16], float dest[16])
{
	int i, j, k, swap;
	float t;
	float temp[4][4];

	memcpy(temp, m, 16*sizeof(float));

	Identity(dest);

	for (i = 0; i < 4; i++) {
	/*
	** Look for largest element in column
	*/
	swap = i;
	for (j = i + 1; j < 4; j++) {
		if (fabs(temp[j][i]) > fabs(temp[i][i])) {
		swap = j;
		}
	}

	if (swap != i) {
		/*
		** Swap rows.
		*/
		for (k = 0; k < 4; k++) {
		t = temp[i][k];
		temp[i][k] = temp[swap][k];
		temp[swap][k] = t;

		t = dest[i*4+k];
		dest[i*4+k] = dest[swap*4+k];
		dest[swap*4+k] = t;
		}
	}

	if (temp[i][i] == 0) {
		/*
		** No non-zero pivot.  The matrix is singular, which shouldn't
		** happen.  This means the user gave us a bad matrix.
		*/
		return false;
	}

	t = temp[i][i];
	for (k = 0; k < 4; k++) {
		temp[i][k] /= t;
		dest[i*4+k] /= t;
	}
	for (j = 0; j < 4; j++) {
		if (j != i) {
		t = temp[j][i];
		for (k = 0; k < 4; k++) {
			temp[j][k] -= temp[i][k]*t;
			dest[j*4+k] -= dest[i*4+k]*t;
		}
		}
	}
	}
	return true;
}

		/// Matrix inversion in place
		/**
		\param m [in, out] Matrix to invert
		\return True if inversion has succeeded
		*/
		inline bool InvertIP(float m[16]) {return Invert(m, m);}

		/// Compute 'Unprojection' matrix
		/**
		\param projMatrix [in] Projection matrix
		\param modelviewMatrix [in] Modelview matrix
		\param unprojectMatrix [out] (projMatrix*modelviewMatrix)^(-1)
		\return Pointer on 'unprojectMatrix'
		*/
		inline float *UnProjectMatrix(const float projMatrix[16], const float modelviewMatrix[16], float unprojectMatrix[16]);

		/// Unproject a point in window coordinates
		/**
		\param viewport [in] Viewport parameters
		\param unprjMatrix [in] 'Unprojection' matrix (see above)
		\param winpos [in] Point window coordinates
		\param worldpos [out] Point world coordinates
		\return Pointer on 'worldpos'
		*/
		inline float *UnProject(const int viewport[4], const float unprjMatrix[16], const float winpos[3], float worldpos[3]);

    /// Project a point given in space coordinates
    /**
    \param worldpos [in] Point world coordinates
    \param modelViewMatrix [in] ModelView matrix (if NULL, supposed to be identity)
    \param projMatrix [in] Projection matrix 
		\param viewport [in] Viewport parameters
		\param scrnPos [out] Corresponding screen position
		\return True if succeeded
    
    */
    inline bool Project(const float worldpos[3], const float modelViewMatrix[16], const float projMatrix[16], const int viewport[4], float scrnpos[3]);
    
    /// Generate a 'gluLookAt'-like matrix 
   

	};
  ///\}

#endif //__MATRIX_H__