////////////////////////////////////
// raymarching.cpp
// Raymarching-based volume renderer
// Original coding by Pascal Gautron, February 2007
////////////////////////////////////



#include <raymarching.h>
using namespace std;


CRayMarching::CRayMarching()
:_shader(NULL), _threshold(0.5f), _frameCounter(0), _fps(0.0f), _infoPanel(true), _start(0), _nSteps(500)
  {}


static void ReadLine(char *buffer, FILE *file)
{
  int i=0;
  char tmp=' ';
  while(fscanf(file, "%c", &tmp) != EOF && tmp != '\n')
  {
    buffer[i++] = tmp;
  }
  buffer[i] = '\0';
}

bool CRayMarching::LoadData(const char *iniFile)
{
  FILE *file = fopen(iniFile, "r");
  if (!file)
  {
    STATUS << "Configuration file " << iniFile << "not found";
    return false;
  }
  char fileBase[1024];
  char ext[4];
  char fileCount[10];
  char digits[10];
  char start[10];
  ReadLine(fileBase, file);
  ReadLine(ext, file);
  ReadLine(fileCount, file);
  ReadLine(digits, file);
  ReadLine(start, file);
  
  return LoadData(fileBase, ext, atoi(fileCount), atoi(digits), atoi(start)-1);

}

bool CRayMarching::LoadData(const char *baseName, const char *extension, int numberOfFiles, int digits, int start)
{
  STATUS << "Loading images..." << FLUSH;
  GenFileName(baseName, extension, 1+start, digits);
  ilInit();
  ilImage img;
  if (!img.Load(_tmpFileName))
  {
    WARNING("Image file " << _tmpFileName << " cannot be opened");
    return false;
  }



  int width = img.Width(), height = img.Height(), depth = numberOfFiles;
  _depth = numberOfFiles;
  _3dTex.Init(img.Width(), img.Height(), numberOfFiles, 1);

  _3dTex.SetZSlice(0, img.GetData());

  

  for(int i=2; i <= numberOfFiles; i++)
  {
    GenFileName(baseName, extension, i+start, digits);
    if (!img.Load(_tmpFileName))
    {
      WARNING("Image file " << _tmpFileName << " cannot be opened");
      //return false;
      continue;
    }
    _3dTex.SetZSlice(i-1, img.GetData());
  }
  STATUS << "OK" << ENDL;
  STATUS << "Filling GL Texture..." << FLUSH;
  glGenTextures(1, &_texID);
  glBindTexture(GL_TEXTURE_3D, _texID);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage3D(
      GL_TEXTURE_3D,
      0,
      GL_ALPHA,
      img.Width(),
      img.Height(),
      numberOfFiles,
      0,
      GL_ALPHA,
      GL_UNSIGNED_BYTE,
      _3dTex.GetData()
      );
    STATUS << "OK" << ENDL;
 
  ComputeGradientsCPU(width, height, depth);




    STATUS << "Compiling shader..." << FLUSH;
    if (_shader==NULL)
    {
      _shader = new CShader("./Shaders/raymarching.vertex", "./Shaders/raymarching.fragment");
      _shader->Activate();
      _shader->SetIntUniform("volumeTex", 0);
      _shader->SetIntUniform("normalTex", 1);
    }
    STATUS << "OK" << ENDL;
  return true;
}

void CRayMarching::ComputeGradientsCPU(int width, int height, int depth)
{
    STATUS << "Computing gradients... " << FLUSH;
    CPerfTimer gradTime(TRUE);
  _normalTex.Init(width, height, depth, 3);

  unsigned char normalUB[3];
  int distance = 1;
  float weight = 1.0f/(2.0f*distance*255.0f);
    for(int z = distance; z < depth-distance; z++)
    {
      for(int y=distance; y < height-distance; y++)
      {
        for(int x=distance; x < width-distance; x++)
        {
          float normal[3] = {0.0f, 0.0f, 0.0f};
          for(int d=-distance; d < 0; d++)
          {
            unsigned char *xOp = _3dTex.GetPixel(x+d, y, z);
            unsigned char *yOp = _3dTex.GetPixel(x, y+d, z);
            unsigned char *zOp = _3dTex.GetPixel(x, y, z+d);
            normal[0] += -weight*(*xOp);
            normal[1] += -weight*(*yOp);
            normal[2] += -weight*(*zOp);
          }
          for(int d=1; d <= distance; d++)
          {
            unsigned char *xOp = _3dTex.GetPixel(x+d, y, z);
            unsigned char *yOp = _3dTex.GetPixel(x, y+d, z);
            unsigned char *zOp = _3dTex.GetPixel(x, y, z+d);
            normal[0] += weight*(*xOp);
            normal[1] += weight*(*yOp);
            normal[2] += weight*(*zOp);
          }

          float len = InvSqrt(
            normal[0]*normal[0]+
            normal[1]*normal[1]+
            normal[2]*normal[2]);
          normal[0]*= len;
          normal[1]*= len;
          normal[2]*= len;

          normal[0] = (normal[0]+1.0f)*0.5f;
          normal[1] = (normal[1]+1.0f)*0.5f;
          normal[2] = (normal[2]+1.0f)*0.5f;

          normalUB[0] = (unsigned char)(255.0f*normal[0]);
          normalUB[1] = (unsigned char)(255.0f*normal[1]);
          normalUB[2] = (unsigned char)(255.0f*normal[2]);

          _normalTex.SetPixel(x, y, z, normalUB);
        }
      }
    }

  for(int z=0; z < distance; z++)
  {
     for(int y=0; y < height; y++)
      {
        for(int x=0; x < width; x++)
        {
          memcpy(_normalTex.GetPixel(x, y, z), _normalTex.GetPixel(x, y, distance), 3*sizeof(unsigned char)); 
          memcpy(_normalTex.GetPixel(x, y, depth-z-1), _normalTex.GetPixel(x, y, depth-distance-1), 3*sizeof(unsigned char)); 
        }
      }
  }

  for(int z=0; z < depth; z++)
  {
     for(int y=0; y < distance; y++)
      {
        for(int x=0; x < width; x++)
        {
          memcpy(_normalTex.GetPixel(x, y, z), _normalTex.GetPixel(x, distance, z), 3*sizeof(unsigned char)); 
          memcpy(_normalTex.GetPixel(x, height-y-1, z), _normalTex.GetPixel(x, height-distance-1, z), 3*sizeof(unsigned char)); 
        }
      }
  }
  for(int z=0; z < depth; z++)
  {
     for(int y=0; y < height; y++)
      {
        for(int x=0; x < distance; x++)
        {
          memcpy(_normalTex.GetPixel(x, y, z), _normalTex.GetPixel(distance, y, z), 3*sizeof(unsigned char)); 
          memcpy(_normalTex.GetPixel(width-x-1, y, z), _normalTex.GetPixel(width-distance-1, y, z), 3*sizeof(unsigned char)); 
        }
      }
  }

  gradTime.Stop();

  glGenTextures(1, &_normalTexID);
  glBindTexture(GL_TEXTURE_3D, _normalTexID);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage3D(
      GL_TEXTURE_3D,
      0,
      GL_RGB,
      width,
      height,
      depth,
      0,
      GL_RGB,
      GL_UNSIGNED_BYTE,
      _normalTex.GetData()
      );

    STATUS << "OK (" << gradTime.Elapsedms() << "ms)" << ENDL;
}


const char *CRayMarching::GenFileName(const char *baseName, const char *extension, int id, int digits)
{
  sprintf(_tmpFileName, "%s", baseName);
  int pos = (int)strlen(baseName);
  

  int tmpid = id;
  int iddigits=0;
  while(tmpid > 0)
  {
    iddigits++;
    tmpid/= 10;
  }
  iddigits = max(1, iddigits);
  int zeroes = digits-iddigits;
  for(int i=0;i < zeroes; i++)
    _tmpFileName[pos+i] = '0';
  sprintf(_tmpFileName+pos+i, "%d.%s", id, extension);
  return _tmpFileName;
}

void CRayMarching::Render()
{

  _timer.Start();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);



	// Activate face culling
	glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  //glEnable(GL_BLEND);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	// Enable the shaders
  _shader->Activate();

  glActiveTexture(GL_TEXTURE1);
  glEnable(GL_TEXTURE_3D);
  glBindTexture(GL_TEXTURE_3D, _normalTexID);

	glActiveTexture(GL_TEXTURE0);
  // Bind the 3D texture
	glBindTexture(GL_TEXTURE_3D, _texID);

  _shader->Activate();
  _shader->SetFloatUniform("threshold", _threshold);
  _shader->SetIntUniform("nSteps", _nSteps);
	// Draw a cube
	glBegin(GL_QUADS);

		// X-
		glTexCoord3f(0.0f, 0.0f, 0.0f);
		glVertex3f(-1.0f, -1.0f, -1.0f);
		glTexCoord3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-1.0f, -1.0f,  1.0f);
		glTexCoord3f(0.0f, 1.0f, 1.0f);
		glVertex3f(-1.0f,  1.0f,  1.0f);
		glTexCoord3f(0.0f, 1.0f, 0.0f);
		glVertex3f(-1.0f,  1.0f, -1.0f);

		// X+
		glTexCoord3f(1.0f, 0.0f, 1.0f);
		glVertex3f( 1.0f, -1.0f,  1.0f);
		glTexCoord3f(1.0f, 0.0f, 0.0f);
		glVertex3f( 1.0f, -1.0f, -1.0f);
		glTexCoord3f(1.0f, 1.0f, 0.0f);
		glVertex3f( 1.0f,  1.0f, -1.0f);
		glTexCoord3f(1.0f, 1.0f, 1.0f);
		glVertex3f( 1.0f,  1.0f,  1.0f);

		// Y-
		glTexCoord3f(1.0f, 0.0f, 1.0f);
		glVertex3f( 1.0f, -1.0f,  1.0f);
		glTexCoord3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-1.0f, -1.0f,  1.0f);
		glTexCoord3f(0.0f, 0.0f, 0.0f);
		glVertex3f(-1.0f, -1.0f, -1.0f);
		glTexCoord3f(1.0f, 0.0f, 0.0f);
		glVertex3f( 1.0f, -1.0f, -1.0f);

		// Y+
		glTexCoord3f(0.0f, 1.0f, 1.0f);
		glVertex3f(-1.0f,  1.0f,  1.0f);
		glTexCoord3f(1.0f, 1.0f, 1.0f);
		glVertex3f( 1.0f,  1.0f,  1.0f);
		glTexCoord3f(1.0f, 1.0f, 0.0f);
		glVertex3f( 1.0f,  1.0f, -1.0f);
		glTexCoord3f(0.0f, 1.0f, 0.0f);
		glVertex3f(-1.0f,  1.0f, -1.0f);

		// Z-
		glTexCoord3f(1.0f, 0.0f, 0.0f);
		glVertex3f( 1.0f, -1.0f, -1.0f);
		glTexCoord3f(0.0f, 0.0f, 0.0f);
		glVertex3f(-1.0f, -1.0f, -1.0f);
		glTexCoord3f(0.0f, 1.0f, 0.0f);
		glVertex3f(-1.0f,  1.0f, -1.0f);
		glTexCoord3f(1.0f, 1.0f, 0.0f);
		glVertex3f( 1.0f,  1.0f, -1.0f);

		// Z+
		glTexCoord3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-1.0f, -1.0f,  1.0f);
		glTexCoord3f(1.0f, 0.0f, 1.0f);
		glVertex3f( 1.0f, -1.0f,  1.0f);
		glTexCoord3f(1.0f, 1.0f, 1.0f);
		glVertex3f( 1.0f,  1.0f,  1.0f);
		glTexCoord3f(0.0f, 1.0f, 1.0f);
		glVertex3f(-1.0f,  1.0f,  1.0f);

	glEnd();


  glFinish();

 
  _timer.Stop();
  _frameCounter++;

  if (_timer.Elapsed() >= 1.0)
  {
    _fps = _frameCounter/ _timer.Elapsed();
    _frameCounter = 0;
    _timer -= _timer;
  }
  _shader->UseProgram(0);
  DisplayInfoPanel();
  glutPostRedisplay();
  }

void CRayMarching::DisplayInfoPanel()
{
if (_infoPanel)
  {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBindTexture(GL_TEXTURE_2D, 0);
    glBindTexture(GL_TEXTURE_3D, 0);
    glColor4f(0.0f, 0.75f, 0.0f, 0.7f);

    glLineWidth(3.0f);
    
    glPolygonMode(GL_FRONT, GL_FILL);
    glBegin(GL_QUADS);
    glVertex2f(-0.95f, -0.60f);
    glVertex2f(-0.95f, -0.98f);
    glVertex2f(-0.2f, -0.98f);
    glVertex2f(-0.2f, -0.60f);
    glEnd();

    glColor4f(0.0f, 0.0f, 0.0f, 1.0f);
    glPolygonMode(GL_FRONT, GL_LINE);
    glBegin(GL_QUADS);
    glVertex2f(-0.95f, -0.60f);
    glVertex2f(-0.95f, -0.98f);
    glVertex2f(-0.2f, -0.98f);
    glVertex2f(-0.2f, -0.60f);
    glEnd();
    glPolygonMode(GL_FRONT, GL_FILL);

    glColor3f(1.0f, 1.0f, 1.0f);
    float position[] = {-0.92f, -0.65f};


    glDisable(GL_BLEND);

    char buffer[255];
    sprintf(buffer, "Opacity (A/Q): %3.2f", _threshold);
    position[1] -= 0.05f; 
    BitmapOutput2D(position, buffer);

    position[1] -= 0.05f;
    sprintf(buffer, "Steps (Z/S): %d", _nSteps);
    BitmapOutput2D(position, buffer);


    position[1] -= 0.05f;
    sprintf(buffer, "Start slice (E/D): %d", _start);
    BitmapOutput2D(position, buffer);

    position[1] -= 0.05f;
    sprintf(buffer, "FPS: %3.2f", _fps);
    BitmapOutput2D(position, buffer);

    position[1] -= 0.05f;
    sprintf(buffer, "Arrow keys: Move light");
    BitmapOutput2D(position, buffer);

    position[1] -= 0.05f;
    sprintf(buffer, "Space: Toogle info");
    BitmapOutput2D(position, buffer);
  }
}



void CRayMarching::Menu(int item)
{
  glutPostRedisplay();
}
void CRayMarching::CreateMenus()
{
}

void CRayMarching::Key(unsigned char key, int x, int y)
  {
    switch (key)
    {
    case 'a':
    case 'A':
      if (_threshold < 1.0f)
      {
        _threshold += 0.01f;
      }
      break;
    case 'q':
    case 'Q':
      if (_threshold > 0.01f)
      {
        _threshold -= 0.01f;
      }
      break;
    case ' ':
      _infoPanel = !_infoPanel;
    
    case 'E':
    case 'e':
      if (_start < _depth-1)
      {
        _start++;
        
      }
      break;
    case 'D':
    case 'd':
      if (_start > 0 )
      {
        _start--;
      }
      break;

    case 'S':
    case 's':
      if (_nSteps > 10 )
      {
        _nSteps-=10;
      }
      break;
    case 'Z':
    case 'z':
      if (_nSteps < 500)
        _nSteps+=10;
      break;


    }
  }
